﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using LibApp;

namespace Cmat_Elasticity
{
    class Program
    {
        static void Main(string[] args)
        {
            // *********** Start the session ***********
            // Assembly number 8000 is displayed in the main package and should only be set for debugging purposes.
            // Once installed/called directly from the main package it will be overruled automatically.
            // *******************************************************************************
            MatchIDDataSession dataSession = new MatchIDDataSession("8000");

            //************Indicate Dataset   *************
            // indicate on what dataset you want to operate. For the VFM there is only one dataset and as such it will not be called
            // For the main result package there might be multiple datasets to operate on
            // *******************************************************************************
            string SetID = dataSession.UI_SelectDataset("Please select a set from the tree");

            //************Open the Source Dataset   *************
            MatchIDData DataFromMatchID = new MatchIDData(dataSession, SetID);

            // ************Create the Destination Dataset   *************
            MatchIDData DataToMatchID = DataFromMatchID.GetDestinationDataSet_CMAT(dataSession, SetID);

            // ************Retrieve time and space density *************
            int iNumberOfPoints = DataFromMatchID.NumberOfDataPoints;
            int iNumberOfTimeSteps = DataFromMatchID.NumberOfTimeSteps;
          
            // ************Memory allocation for data storage in script *************
            // Both strain and stress memory is allocated
            // *******************************************************************************
            double[,] Exx = new double[iNumberOfTimeSteps, iNumberOfPoints];
            double[,] Eyy = new double[iNumberOfTimeSteps, iNumberOfPoints];
            double[,] Exy = new double[iNumberOfTimeSteps, iNumberOfPoints];
            double[,] Sxx = new double[iNumberOfTimeSteps, iNumberOfPoints];
            double[,] Syy = new double[iNumberOfTimeSteps, iNumberOfPoints];
            double[,] Sxy = new double[iNumberOfTimeSteps, iNumberOfPoints];

            // ************Retrieve the strain variables from the Source dataset *************
            // these variables only need to be recovered once since they do not change withing the optimization
            // Following conventions are adopted:
            // 1. strains are LOGARITHMIC
            // 2. strains are expressed in a COROTATIONAL frame
            // 3. TENSORIAL convention for the shear components
            // *******************************************************************************
            GetStrainDataFromMatchID(DataFromMatchID, iNumberOfPoints, iNumberOfTimeSteps, Exx, Eyy, Exy);

            // ************Retrieve the deformation gradient from the Source dataset *************
            // In case one opts not to work with strain values one can directly operate on the deformation gradient F, which
            // is a 3x3 tensor. In case of a 2D DIC setup only the upperleft 2x2 matrix is complete
            // *************************************************************************************
            //GetDeformationGradientFromMatchID(DataFromMatchID, iNumberOfPoints, iNumberOfTimeSteps);

            // ************Retrieve ProcessValues: Values that are fixed in time and space*************
            // If relevant, Process variables that do not change per iteration and are fixed in time and space
            // can be retrieved. The string naming declaration should be consistent with the label indicated in the 
            // custom material screen
            // *******************************************************************************
            // double Temperature = Convert.ToDouble(DataFromMatchID.GetProcessValue("Temp"));


            // ************Retrieve ConstantValues: Values that change temporal but not spatial*************
            // If relevant: Time, ... can be retrieved 
            // *******************************************************************************
           // GetSpatialConstantValuesFromMatchID(DataFromMatchID, iNumberOfTimeSteps);


            // ************Actual Stress Reconstruction *************
            // CMAT within the VFM is an iterative procedure. Hence, algorithm is contained within
            //  a do{}while(dataSession.Optimisation_WaitForNextIteration()) loop whereby the while
            // argument indicates the script will await for a trigger invoked via the VFM
            // *******************************************************************************
            do
            {
                /* STEP 1*/
                // declaration of the parameters that are involved and that vary per iteration. 
                // the length and string naming declaration should be consistent with the label indicated in the 
                // custom material screen
                double[] dParams = new double[2];
                dParams[0] = Convert.ToDouble(DataFromMatchID.GetProcessValue("Ec"));
                dParams[1] = Convert.ToDouble(DataFromMatchID.GetProcessValue("vc"));

                /* STEP 2*/
                // Actual strain to stress conversion
                // Material model should return CAUCHY stresses
                ElasticityModel(DataToMatchID, dParams, iNumberOfPoints, iNumberOfTimeSteps, Exx, Eyy, Exy, Sxx,Syy,Sxy);

                /* STEP 3*/
                // Send the stress data to MatchID. Two options available
                // Option 1: memory saving but slow
                // SendStressToMatchID_Direct(DataToMatchID, iNumberOfPoints, iNumberOfTimeSteps, Sxx, Syy, Sxy);
                // Option 2: bulk per time step. Much faster
                SendStressToMatchID_BULK(DataToMatchID, iNumberOfPoints, iNumberOfTimeSteps, Sxx, Syy, Sxy);

                /* STEP 4*/
                // Make the assembly aware the current stress conversion is finished
                DataToMatchID.SendToMatchID();
            }
            while (dataSession.Optimisation_WaitForNextIteration());


        }


        private static void ElasticityModel(MatchIDData DataToMatchID, double[] dParams, int iNumberOfPoints, int iNumberOfTimeSteps,
          double[,] Exx, double[,] Eyy, double[,] Exy, double[,] Sxx, double[,] Syy, double[,] Sxy)
        {
            // declaration of useful factors for the elastic material script
            double f1 = dParams[0] / (1 - dParams[1] * dParams[1]); // E/(1-v^2)
            double f2 = dParams[0] / (1 + dParams[1]); // E/(1+v)

            for (int indexPoint = 0; indexPoint < iNumberOfPoints; indexPoint++)
            {
                for (int indexTime = 0; indexTime < iNumberOfTimeSteps; indexTime++)
                {

                    Sxx[indexTime, indexPoint] = f1 * (Exx[indexTime, indexPoint] + dParams[1] * Eyy[indexTime, indexPoint]);
                    Syy[indexTime, indexPoint] = f1 * (Eyy[indexTime, indexPoint] + dParams[1] * Exx[indexTime, indexPoint]);
                    Sxy[indexTime, indexPoint] = f2 * Exy[indexTime, indexPoint];
                }
            }
        }

        private static void GetStrainDataFromMatchID(MatchIDData DataFromMatchID, int iNumberOfPoints, int iNumberOfTimeSteps,
            double[,] Exx, double[,] Eyy, double[,] Exy)
        {
            // ************Retrieve the relevant data indices *************
            int iExxIndex = DataFromMatchID.GetVariableIndex("exx");
            int iEyyIndex = DataFromMatchID.GetVariableIndex("eyy");
            int iExyIndex = DataFromMatchID.GetVariableIndex("exy");


            for (int indexTime = 0; indexTime < iNumberOfTimeSteps; indexTime++)
            {
                for (int indexPoint = 0; indexPoint < iNumberOfPoints; indexPoint++)
                {
                    // Get the relevant data from the Source Dataset
                    Exx[indexTime, indexPoint] = DataFromMatchID.GetValue(indexPoint, indexTime, iExxIndex);
                    Eyy[indexTime, indexPoint] = DataFromMatchID.GetValue(indexPoint, indexTime, iEyyIndex);
                    Exy[indexTime, indexPoint] = DataFromMatchID.GetValue(indexPoint, indexTime, iExyIndex);
                }
            }

        }

        private static void SendStressToMatchID_BULK(MatchIDData DataToMatchID, int iNumberOfPoints, int iNumberOfTimeSteps,
           double[,] Sxx, double[,] Syy, double[,] Sxy)
        {
            // indices of variables to store
            int iSxxIndex = DataToMatchID.GetVariableIndex("sxx");
            int iSyyIndex = DataToMatchID.GetVariableIndex("syy");
            int iSxyIndex = DataToMatchID.GetVariableIndex("sxy");
           // int if22Index = DataToMatchID.GetVariableIndex("f22");

            for (int indexPoint = 0; indexPoint < iNumberOfPoints; indexPoint++)
            {
                double[] Stresses = new double[iNumberOfTimeSteps * 3];
                for (int indexTime = 0; indexTime < iNumberOfTimeSteps; indexTime++)
                {
                    Stresses[indexTime * 3] = Sxx[indexTime, indexPoint];
                    Stresses[indexTime * 3 + 1] = Syy[indexTime, indexPoint];
                    Stresses[indexTime * 3 + 2] = Sxy[indexTime, indexPoint];
                 //   DataToMatchID.SetValue(indexPoint, indexTime, if22Index, 0.5);
                }
                // Faster procedure: only once over all time step. This avoids continuous communication
                DataToMatchID.SetVector_byArray(indexPoint, iSxxIndex, iSyyIndex, iSxyIndex, Stresses);
              
            }

        }

        private static void SendStressToMatchID_Direct(MatchIDData DataToMatchID, int iNumberOfPoints, int iNumberOfTimeSteps,
          double[,] Sxx, double[,] Syy, double[,] Sxy)
        {
            // indices of variables to store
            int iSxxIndex = DataToMatchID.GetVariableIndex("sxx");
            int iSyyIndex = DataToMatchID.GetVariableIndex("syy");
            int iSxyIndex = DataToMatchID.GetVariableIndex("sxy");

            for (int indexPoint = 0; indexPoint < iNumberOfPoints; indexPoint++)
            {
                for (int indexTime = 0; indexTime < iNumberOfTimeSteps; indexTime++)
                {
                    // Set the relevant data to the Destination Dataset. Individual one by one. 
                    // Does not require data storage but much slower
                    DataToMatchID.SetValue(indexPoint, indexTime, iSxxIndex, Sxx[indexTime, indexPoint]);
                    DataToMatchID.SetValue(indexPoint, indexTime, iSyyIndex, Syy[indexTime, indexPoint]);
                    DataToMatchID.SetValue(indexPoint, indexTime, iSxyIndex, Sxy[indexTime, indexPoint]);
                }
            }

        }

        private static void GetDeformationGradientFromMatchID(MatchIDData DataFromMatchID, int iNumberOfPoints, int iNumberOfTimeSteps)
        {
            int if00Index = DataFromMatchID.GetVariableIndex("f00");
            int if01Index = DataFromMatchID.GetVariableIndex("f01");
            int if02Index = DataFromMatchID.GetVariableIndex("f02");
            int if10Index = DataFromMatchID.GetVariableIndex("f10");
            int if11Index = DataFromMatchID.GetVariableIndex("f11");
            int if12Index = DataFromMatchID.GetVariableIndex("f12");
            int if20Index = DataFromMatchID.GetVariableIndex("f20");
            int if21Index = DataFromMatchID.GetVariableIndex("f21");
            int if22Index = DataFromMatchID.GetVariableIndex("f22");

            double[,][,] F = new double[iNumberOfTimeSteps, iNumberOfPoints][,];

            for (int indexTime = 0; indexTime < iNumberOfTimeSteps; indexTime++)
            {
                for (int indexPoint = 0; indexPoint < iNumberOfPoints; indexPoint++)
                {
                    F[indexTime, indexPoint] = new double[3, 3];
                    F[indexTime,indexPoint][0,0] = DataFromMatchID.GetValue(indexPoint, indexTime, if00Index);
                    F[indexTime, indexPoint][0, 1] = DataFromMatchID.GetValue(indexPoint, indexTime, if01Index);
                    F[indexTime, indexPoint][0, 2] = DataFromMatchID.GetValue(indexPoint, indexTime, if02Index);
                    F[indexTime, indexPoint][1, 0] = DataFromMatchID.GetValue(indexPoint, indexTime, if10Index);
                    F[indexTime, indexPoint][1, 1] = DataFromMatchID.GetValue(indexPoint, indexTime, if11Index);
                    F[indexTime, indexPoint][1, 2] = DataFromMatchID.GetValue(indexPoint, indexTime, if12Index);
                    F[indexTime, indexPoint][2, 0] = DataFromMatchID.GetValue(indexPoint, indexTime, if20Index);
                    F[indexTime, indexPoint][2, 1] = DataFromMatchID.GetValue(indexPoint, indexTime, if21Index);
                    F[indexTime, indexPoint][2, 2] = DataFromMatchID.GetValue(indexPoint, indexTime, if22Index);
                }
            }

        }

        private static void GetSpatialConstantValuesFromMatchID(MatchIDData DataFromMatchID, int iNumberOfTimeSteps)
        {
            int iTimeIndex = DataFromMatchID.GetVariableIndex("time");
            double[] Time = new double[iNumberOfTimeSteps];
            for (int i = 0; i < iNumberOfTimeSteps; i++)
            {
                Time[i] = DataFromMatchID.GetConstantValue(i, iTimeIndex);
            }
        }

    }
}
