# LibApp can be found in Appstore's main directory
import sys, os
dir = os.path.dirname(__file__)
sys.path.append(os.path.realpath(os.path.join(dir, '..\..\Libs')))
import clr
clr.AddReference(r'LibApp')

import LibApp
from LibApp import MatchIDDataSession
from LibApp import MatchIDDataSelection
from LibApp import MatchIDData



#Main script			
def Main():			
	# *********** Start the session ***********
	# Assembly number 8000 is displayed in the main package and should only be set for debugging purposes.
	# Once installed/called directly from the main package it will be overruled automatically.
	#*******************************************************************************

	dataSession = MatchIDDataSession("8000")

	# ************Indicate Dataset   *************
	# indicate on what dataset you want to operate. For the VFM there is only one dataset and as such it will not be called
	# For the main result package there might be multiple datasets to operate on
	# *******************************************************************************
	 
	SetID = dataSession.UI_SelectDataset("Please select a set from the tree")
	 
	#************Open the Source Dataset   *************
	DataFromMatchID = MatchIDData(dataSession, SetID)

	#************Create the Destination Dataset   *************
	DataToMatchID = DataFromMatchID.GetDestinationDataSet_CMAT(dataSession, SetID);

	#************Retrieve time and space density *************
	iNumberOfPoints = DataFromMatchID.NumberOfDataPoints;
	iNumberOfTimeSteps = DataFromMatchID.NumberOfTimeSteps;

	#************Memory allocation for data storage in script *************
	# Both strain and stress memory is allocated
	# *******************************************************************************
	Exx = [0] * iNumberOfTimeSteps
	Eyy = [0] * iNumberOfTimeSteps
	Exy = [0] * iNumberOfTimeSteps
	Sxx = [0] * iNumberOfTimeSteps
	Syy = [0] * iNumberOfTimeSteps
	Sxy = [0] * iNumberOfTimeSteps
	for i in range(iNumberOfTimeSteps):
		Exx[i] = [0] * iNumberOfPoints
		Eyy[i] = [0] * iNumberOfPoints
		Exy[i] = [0] * iNumberOfPoints
		Sxx[i] = [0] * iNumberOfPoints
		Syy[i] = [0] * iNumberOfPoints
		Sxy[i] = [0] * iNumberOfPoints
		
	# ************Retrieve the strain variables from the Source dataset *************
	# these variables only need to be recovered once since they do not change withing the optimization
	# Following conventions are adopted:
	# 1. strains are LOGARITHMIC
	# 2. strains are expressed in a COROTATIONAL frame
	# 3. TENSORIAL convention for the shear components
	# *******************************************************************************

	GetStrainDataFromMatchID(DataFromMatchID, iNumberOfPoints, iNumberOfTimeSteps, Exx, Eyy, Exy)
		
		
	#************Retrieve the deformation gradient from the Source dataset *************
	# In case one opts not to work with strain values one can directly operate on the deformation gradient F, which
	# is a 3x3 tensor. In case of a 2D DIC setup only the upperleft 2x2 matrix is complete
	# *************************************************************************************
	
	#GetDeformationGradientFromMatchID(DataFromMatchID, iNumberOfPoints, iNumberOfTimeSteps)
	
	#************Retrieve ProcessValues: Values that are fixed in time and space*************
	#If relevant, Process variables that do not change per iteration and are fixed in time and space
	# can be retrieved. The string naming declaration should be consistent with the label indicated in the 
	# custom material screen
	# *******************************************************************************
	
	#Temperature = float(DataFromMatchID.GetProcessValue("Temp"))

	#************Retrieve ConstantValues: Values that change temporal but not spatial*************
	#If relevant: Time, ... can be retrieved 
	#******************************************************************************* 
	
	#GetSpatialConstantValuesFromMatchID(DataFromMatchID, iNumberOfTimeSteps)
	
	
	#************Actual Stress Reconstruction *************
    # CMAT within the VFM is an iterative procedure. Hence, algorithm is contained within
    # a while(dataSession.Optimisation_WaitForNextIteration()) loop whereby the 
	# the script will await for a trigger invoked via the VFM
    # *******************************************************************************
	while True:
	
		#STEP 1
		# declaration of the parameters that are involved and that vary per iteration. 
		# the length and string naming declaration should be consistent with the label indicated in the 
		# custom material screen
		dParams = [0]*2;
		dParams[0] = float(DataFromMatchID.GetProcessValue("Ec"))
		dParams[1] = float(DataFromMatchID.GetProcessValue("vc"))
		
		#STEP 2
		#Actual strain to stress conversion
		# Material model should return CAUCHY stresses
		ElasticityModel(DataToMatchID, dParams, iNumberOfPoints, iNumberOfTimeSteps, Exx, Eyy, Exy, Sxx,Syy,Sxy)

		#STEP 3
		#Send the stress data to MatchID. Two options available
		#Option 1: memory saving but slow
		#SendStressToMatchID_Direct(DataToMatchID, iNumberOfPoints, iNumberOfTimeSteps, Sxx, Syy, Sxy);
		# Option 2: bulk per time step. Much faster
		SendStressToMatchID_BULK(DataToMatchID, iNumberOfPoints, iNumberOfTimeSteps, Sxx, Syy, Sxy)
		
		#STEP 4
		#Make the assembly aware the current stress conversion is finished
		DataToMatchID.SendToMatchID();

		#STEP 5
		#await for trigger from MatchID
		isNextTrigger = dataSession.Optimisation_WaitForNextIteration()
		
		#STEP 6
		#check if next trigger is valid or not
		if isNextTrigger is False:
			break
			
def GetStrainDataFromMatchID(DataFromMatchID,iNumberOfPoints,iNumberOfTimeSteps,Exx,Eyy,Exy):
	# ************Retrieve the relevant data indices *************
	iExxIndex = DataFromMatchID.GetVariableIndex("exx")
	iEyyIndex = DataFromMatchID.GetVariableIndex("eyy")
	iExyIndex = DataFromMatchID.GetVariableIndex("exy")
	for indexTime in range(iNumberOfTimeSteps):
		for indexPoint in range(iNumberOfPoints):
			#Get the relevant data from the Source Dataset
			Exx[indexTime][indexPoint] =  DataFromMatchID.GetValue(indexPoint, indexTime, iExxIndex)
			Eyy[indexTime][indexPoint] =  DataFromMatchID.GetValue(indexPoint, indexTime, iEyyIndex)
			Exy[indexTime][indexPoint] =  DataFromMatchID.GetValue(indexPoint, indexTime, iExyIndex)


def ElasticityModel(DataToMatchID, dParams, iNumberOfPoints, iNumberOfTimeSteps, Exx, Eyy, Exy, Sxx,Syy,Sxy):
	# declaration of useful factors for the elastic material script
	f1 = dParams[0] / (1 - dParams[1] * dParams[1]); # E/(1-v^2)
	f2 = dParams[0] / (1 + dParams[1]); # E/(1+v)

	for indexTime in range(iNumberOfTimeSteps):
		for indexPoint in range(iNumberOfPoints):
			Sxx[indexTime][indexPoint] = f1 * (Exx[indexTime][indexPoint] + dParams[1] * Eyy[indexTime][indexPoint])
			Syy[indexTime][indexPoint] = f1 * (Eyy[indexTime][indexPoint] + dParams[1] * Exx[indexTime][indexPoint])
			Sxy[indexTime][indexPoint] = f2 * Exy[indexTime][indexPoint]
                		
def SendStressToMatchID_BULK(DataToMatchID, iNumberOfPoints, iNumberOfTimeSteps, Sxx, Syy, Sxy):
	#indices of variables to store
	iSxxIndex = DataToMatchID.GetVariableIndex("sxx")
	iSyyIndex = DataToMatchID.GetVariableIndex("syy")
	iSxyIndex = DataToMatchID.GetVariableIndex("sxy")
	
	for indexPoint in range(iNumberOfPoints):
		Stresses = [0]*(iNumberOfTimeSteps * 3)
		for indexTime in range(iNumberOfTimeSteps):
			Stresses[(indexTime) * 3] = Sxx[indexTime][indexPoint];
			Stresses[(indexTime) * 3 + 1] = Syy[indexTime][indexPoint];
			Stresses[(indexTime) * 3 + 2] = Sxy[indexTime][indexPoint];
			
			# Faster procedure: only once over all time step. This avoids continuous communication
			DataToMatchID.SetVector_byArray(indexPoint, iSxxIndex, iSyyIndex, iSxyIndex, Stresses);

def SendStressToMatchID_Direct(DataToMatchID, iNumberOfPoints, iNumberOfTimeSteps, Sxx, Syy, Sxy):
	#indices of variables to store
	iSxxIndex = DataToMatchID.GetVariableIndex("sxx")
	iSyyIndex = DataToMatchID.GetVariableIndex("syy")
	iSxyIndex = DataToMatchID.GetVariableIndex("sxy")
	
	for indexTime in range(iNumberOfTimeSteps):
		for indexPoint in range(iNumberOfPoints):
			#Set the relevant data to the Destination Dataset. Individual one by one. 
			#Does not require data storage but much slower
			DataToMatchID.SetValue(indexPoint, indexTime, iSxxIndex, Sxx[indexTime][indexPoint])
			DataToMatchID.SetValue(indexPoint, indexTime, iSyyIndex, Syy[indexTime][indexPoint])
			DataToMatchID.SetValue(indexPoint, indexTime, iSxyIndex, Sxy[indexTime][indexPoint])
			
def GetDeformationGradientFromMatchID(DataFromMatchID, iNumberOfPoints, iNumberOfTimeSteps):
	if00Index = DataFromMatchID.GetVariableIndex("f00")
	if01Index = DataFromMatchID.GetVariableIndex("f01")
	if02Index = DataFromMatchID.GetVariableIndex("f02")
	if10Index = DataFromMatchID.GetVariableIndex("f10")
	if11Index = DataFromMatchID.GetVariableIndex("f11")
	if12Index = DataFromMatchID.GetVariableIndex("f12")
	if20Index = DataFromMatchID.GetVariableIndex("f20")
	if21Index = DataFromMatchID.GetVariableIndex("f21")
	if22Index = DataFromMatchID.GetVariableIndex("f22")

	F = [0] * iNumberOfTimeSteps
	for i in range(iNumberOfTimeSteps):
		F[i] = [0] * iNumberOfPoints
		for j in range(iNumberOfPoints):
			F[i][j] = [0]*3
			for k in range(3):
				F[i][j][k] = [0]*3
	for indexTime in range(iNumberOfTimeSteps):
		for indexPoint in range(iNumberOfPoints):
			F[indexTime][indexPoint][0][0] = DataFromMatchID.GetValue(indexPoint, indexTime, if00Index)
			F[indexTime][indexPoint][0][1] = DataFromMatchID.GetValue(indexPoint, indexTime, if01Index)
			F[indexTime][indexPoint][0][2] = DataFromMatchID.GetValue(indexPoint, indexTime, if02Index)
			F[indexTime][indexPoint][1][0] = DataFromMatchID.GetValue(indexPoint, indexTime, if10Index)
			F[indexTime][indexPoint][1][1] = DataFromMatchID.GetValue(indexPoint, indexTime, if11Index)
			F[indexTime][indexPoint][1][2] = DataFromMatchID.GetValue(indexPoint, indexTime, if12Index)
			F[indexTime][indexPoint][2][0] = DataFromMatchID.GetValue(indexPoint, indexTime, if20Index)
			F[indexTime][indexPoint][2][1] = DataFromMatchID.GetValue(indexPoint, indexTime, if21Index)
			F[indexTime][indexPoint][2][2] = DataFromMatchID.GetValue(indexPoint, indexTime, if22Index)

def GetSpatialConstantValuesFromMatchID(DataFromMatchID, iNumberOfTimeSteps):
	iTimeIndex = DataFromMatchID.GetVariableIndex("time")
	
	Time = [0]*iNumberOfTimeSteps
	for indexTime in range(iNumberOfTimeSteps):
		Time[indexTime] = DataFromMatchID.GetConstantValue(indexTime, iTimeIndex)
		print(Time[indexTime])
            
Main()